/* 
 * Copyright (C) 2004 Romain Dolbeau <romain@dolbeau.org>
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */

package lavcGenetic;

import java.util.*;
import java.io.*;
import javax.swing.*;

public class lavcGeneticMain {

    /* number of tests to run initially. each subsequent path run 1/3 that number */
    /* can be changed in UI */
    public static final int kSimultaneousTests = 3 * 5; // must be multiple of 3
    /* source file & non-encoding options */
    public static final String kDefaultSource = 
        "/local/dolbeau/test_TigreDragon.vob -frames 100 -ss 0:02:00 -vf crop=720:422:0:78,scale=576:240";
    /* maximum number of pass (maximum number of run will be (kMaxPass + 2) * (kSimultaneousTests / 3) */
    /* only a safety, as the UI can stop the run */
    public static final int kMaxPass = 1000000;
    /* after that time is spend w/o any activy from mplayer, assume it's dead */
    /* leftover from buggy JVM */
    public static final int kTimeLimit = 15; // seconds
    /* percent chance of randomizing each option */
    /* can be changed in UI */
    public static final int kRandomBreeding = 10;
    /* percentage within best PSNR to look for faster result */
    public static final double kWithinPercentage = 1.0;
    /* max number of attempts when randomizing (in case the generated combinations is already done) */
    public static final int kMaxAttempts = 5;
    /* filename to take the machine list (one hostname per line) */
    public static final String kMachineFilename = "/tmp/machine.list";
    /* unused ATM */
    public static final int kNumBestResults = 3;
    /* default bitrate for encoding */
    /* can be changed in UI */
    public static final int kDefaultBitrate = 650;


    
    // private static Set allCombinations = new HashSet();
    private static lavcResultsPool allResults = new lavcResultsPool();
    private static Random random = new Random(0); // known seed for now, so results are reproductible

    static void err_println(String s) {
        synchronized (System.err) {
            System.err.println(s);
        }
    }

    static void out_println(String s) {
        synchronized (System.out) {
            System.out.println(s);
        }
    }

    private static List getMachineList(String filename) {
        List theSet = new ArrayList();
        try {
             InputStream is = new FileInputStream(filename);
             BufferedReader br = new BufferedReader(new InputStreamReader(is));
             String line = br.readLine();
             while (line != null) {
                 theSet.add(line);
                 line = br.readLine();
             }
             br.close();
             is.close();
        } catch (Exception e) { lavcGeneticMain.err_println("Exception " + e + " while reading machines file"); }
        
        return theSet;
    }

    private static void computeThemAll(List newList, lavcJobDist ljd, lavcResultsPool lrp, lavcCodecData lcd, int bitrate, boolean twopass) {
        int nSimultaneousTests = newList.size();
        Thread[] allThreads = new Thread[nSimultaneousTests];
        Iterator it = newList.iterator();
        int j = 0;
        while (it.hasNext()) {
            lavcOptionList options = (lavcOptionList)it.next();

            //lavcGeneticMain.err_println("TRYING (" + j + "): " + options);
            
            lavcEncodeData led = new lavcEncodeData(options, lcd);
            led.source = kDefaultSource;
            led.workdir = "/tmp/" + led.getLacvOptsString();
            
            allThreads[j] = new lavcTestThread(j, led, ljd, lrp, kTimeLimit, bitrate, twopass);
            
            allThreads[j].start();
            j++;
        }
        for (int i = 0; i < j ; i++) {
            try {
                //lavcGeneticMain.err_println("WAITING FOR: " + i);
                allThreads[i].join();
            } catch (Exception e) { lavcGeneticMain.err_println("Exception " + e + " while waiting for thread #" + i); }
        }
    }

    static Set buildLavcOpts() {
        Set allOptions = new HashSet();

        List v = new ArrayList();

        /* ***** mbd ***** */
        allOptions.add(new lavcOption_Integer("mbd", 2, 2, true));

        /* ***** vme ***** */
        v.clear();
        //v.add(new Integer(1));
        v.add(new Integer(4));
        allOptions.add(new lavcOption_Integer("vme", v, true));

        /* ***** *cmp ***** */
        List cmpList = new ArrayList();
        cmpList.add("cmp");
        cmpList.add("precmp");
        cmpList.add("subcmp");
        v.clear();
        //v.add(new Integer(0));
        v.add(new Integer(1));
        v.add(new Integer(2));
        v.add(new Integer(3));
        //v.add(new Integer(5));
        v.add(new Integer(6));
        allOptions.add(new lavcOption_Integer(cmpList, v, true));

        /* ***** *dia ***** */
        List diaList = new ArrayList();
        diaList.add("dia");
        diaList.add("predia");
        v.clear();
        v.add(new Integer(1));
        //v.add(new Integer(2));
        //v.add(new Integer(-1));
        v.add(new Integer(-2));
        //v.add(new Integer(-3));
        allOptions.add(new lavcOption_Integer(diaList, v, true));

        /* ***** B-frames ***** */
         allOptions.add(new lavcOption_Integer("vmax_b_frames", 0, 2, true));

        /* ***** qpel, mv0, qprd, v4mv ***** */
        allOptions.add(new lavcOption_Flag("qpel", false));
        allOptions.add(new lavcOption_Flag("mv0", false));
        allOptions.add(new lavcOption_Flag("qprd", false));
        allOptions.add(new lavcOption_Flag("v4mv", false));

        /* ***** trell ***** */
        v.clear();
        v.add("trell");
        v.add("trell:cbp");
        v.add("trell:qns=1");
        v.add("trell:qns=2");
        v.add("trell:qns=3");
        v.add("trell:qns=1:cbp");
        v.add("trell:qns=2:cbp");
        v.add("trell:qns=3:cbp");
        allOptions.add(new lavcOption_Flag(v, false));

        
        v.clear();
        v.add(new Integer(-4));
        allOptions.add(new lavcOption_Integer("vlelim",  v, false));
        v.clear();
        v.add(new Integer(7));
        allOptions.add(new lavcOption_Integer("vcelim",  v, false));
        v.clear();
        v.add(new Float(0.1));
        v.add(new Float(0.2));
        allOptions.add(new lavcOption_Float("lumi_mask",  v, false));
        v.clear();
        v.add(new Float(0.1));
        v.add(new Float(0.2));
        allOptions.add(new lavcOption_Float("dark_mask",  v, false));
        v.clear();
        v.add(new Float(0.1));
        v.add(new Float(0.2));
        allOptions.add(new lavcOption_Float("tcplx_mask",  v, false));
        v.clear();
        v.add(new Float(0.1));
        v.add(new Float(0.2));
        allOptions.add(new lavcOption_Float("scplx_mask",  v, false));
        v.clear();
        v.add(new Float(0.1));
        v.add(new Float(0.2));
        allOptions.add(new lavcOption_Float("p_mask", v, false));
        
        lavcGeneticMain.err_println("CONTROL: " + allOptions);

        allOptions.add(new lavcOption_Flag("psnr", true));
        allOptions.add(new lavcOption_Flag("vcodec=mpeg4", true));
        allOptions.add(new lavcOption_Flag("autoaspect", true));

        return allOptions;
    }

    static Set buildXvidencOpts() {
        Set allOptions = new HashSet();

        List v = new ArrayList();

        allOptions.add(new lavcOption_Integer("me_quality", 0, 6, true));

        allOptions.add(new lavcOption_Flag("qpel", false));
        allOptions.add(new lavcOption_Flag("gmc", false));
        allOptions.add(new lavcOption_Flag("hq_ac", false));
        allOptions.add(new lavcOption_Flag("chroma_me", false));
        allOptions.add(new lavcOption_Flag("chroma_opt", false));
        allOptions.add(new lavcOption_Integer("max_bframes", 0, 2, true));
        allOptions.add(new lavcOption_Integer("vhq", 0, 4, true));


        allOptions.add(new lavcOption_Flag("psnr", true));
        allOptions.add(new lavcOption_Flag("autoaspect", true));

        return allOptions;
    }

    public static void main(String[] arg) {

        createNewMainDisplay();

    }

    static int numberOfCombinations(Set allSelectedOptions) {
        int allCombinationsCount = 1;
        Iterator it = allSelectedOptions.iterator();
        while (it.hasNext()) {
            lavcOption o = (lavcOption)it.next();
            List oAll = o.getAllVariants();
            allCombinationsCount = allCombinationsCount * oAll.size();
        }
        return allCombinationsCount;
    }
    
    static boolean done = false;
    static final Object doneO = new Object(); 
    static boolean isDone() {
        boolean temp = false;
        synchronized (doneO) {
            temp = done;
        }
        return temp;
    }
    static void setDone() {
        synchronized (doneO) {
            done = true;;
        }
    }
    

    static void doTheJob(Set allSelectedOptions, lavcAllStatisticsDisplay lasd, lavcCodecData lcd,
                         int nSimultaneousTests, int nRandomBreeding, int bitrate, boolean twopass
                         ) {
        Iterator it;

        int allCombinationsCount = numberOfCombinations(allSelectedOptions);
        
        lavcGeneticMain.err_println("Total options combinations: " + allCombinationsCount);
        
        List machineList = getMachineList(kMachineFilename);

        lavcGeneticMain.err_println("Machines: " + machineList);

        lavcJobDist ljd = new lavcJobDist(machineList);

        Thread[] allThreads = new Thread[nSimultaneousTests];

        /* **** STEP ONE ***** */

        int step = 1;

        // List baseList = new ArrayList(allCombinations);
        List newList = new ArrayList();

        for (int i = 0; i < nSimultaneousTests; i++) {
            lavcOptionList lol = null;
            int j = 0;
            do {
                lol = lavcOptionList.buildRandomList(allSelectedOptions, random);
                j++;
            } while (newList.contains(lol) && (j < kMaxAttempts));
            if (!newList.contains(lol))
                newList.add(lol);
        }

        computeThemAll(newList, ljd, allResults, lcd, bitrate, twopass);
        
        //lavcGeneticMain.err_println("ALL Results (" + step + "): " + allResults.getAllResults());
        lavcGeneticMain.err_println("Results computed so far: " + allResults.numberOfResults() + " (" + (100. * (double)allResults.numberOfResults() / (double)allCombinationsCount) + " % of all)");
        lavcOptionList br = allResults.getBestResultKey();
        lavcGeneticMain.err_println("Best Result (" + step + "): " + br + " [" + allResults.getResult(br) + "]");
        lavcOptionList fwpb = allResults.fasterWithinPercentBestKey(kWithinPercentage);
        lavcGeneticMain.err_println("Faster within " + kWithinPercentage + "% (" + step + "): " + fwpb + " [" + allResults.getResult(fwpb) + "]");
        lavcGeneticMain.err_println("Statistics:\n" + allResults.optionStatistics(allSelectedOptions));

        if (lasd != null) {
            lasd.refresh(allResults, br, step, step, fwpb, kWithinPercentage);
        }

        /* ***** STEP TWO AND MORE ***** */

        int bestStep = step;
        lavcResultsPool allButN = new lavcResultsPool(allResults);

        while (!isDone()) {
            step ++;
            int removeNumber = (nSimultaneousTests * 2 / 3) + (nSimultaneousTests - allButN.numberOfResults());

            allButN = allButN.getAllButHalfTheWorstN(removeNumber, random);

            //lavcGeneticMain.err_println("ALL BUT N (" + validResult * 2 / 3 + "): " + allButN);

            int keepUpFactor = (nSimultaneousTests * 2 / 3) - allButN.numberOfResults();
            
            newList = lavcOptionList.breedThemAll(allButN.getListOfKeys(), (nSimultaneousTests / 3) + keepUpFactor, allResults, nRandomBreeding, allSelectedOptions, random);

            if (newList.size() == 0) { // we failed generating anything, try fully random
                newList = lavcOptionList.breedThemAll(allButN.getListOfKeys(), (nSimultaneousTests / 3) + keepUpFactor, allResults, 100, allSelectedOptions, random);
            }

            lavcGeneticMain.err_println("Step " + step + " starts " + newList.size() + " runs.");
            
            if ((newList.size() == 0) || (step >= kMaxPass))
                setDone();
            
            //lavcGeneticMain.err_println("CHILDREN: " + newList);
            if (newList.size() > 0) {
                computeThemAll(newList, ljd, allButN, lcd, bitrate, twopass);
                allResults.mergeResults(allButN);
            
                //lavcGeneticMain.err_println("ALL Results (" + step + "): " + allResults.getAllResults());
                lavcOptionList br2 = allResults.getBestResultKey();
                boolean unchanged = true;
                if (br2 != null)
                    unchanged = br2.equals(br);
                else
                    setDone();
                if (!unchanged)
                    bestStep = step;
                lavcGeneticMain.err_println("Results computed so far: " + allResults.numberOfResults() + " (" + (100. * (double)allResults.numberOfResults() / (double)allCombinationsCount) + " % of all)");
                lavcGeneticMain.err_println("Best Result (" + step + "): " + br2 + " [" + allResults.getResult(br2) + "]" + 
                                            (unchanged ? " {UNCHANGED since " + bestStep + "}" : " {IMPROVED}"));
                fwpb = allResults.fasterWithinPercentBestKey(kWithinPercentage);
                lavcGeneticMain.err_println("Faster within " + kWithinPercentage + "% (" + step + "): " + fwpb + " [" + allResults.getResult(fwpb) + "]");
                lavcGeneticMain.err_println("Statistics (best only):\n" + allButN.optionStatistics(allSelectedOptions));
                
                if (lasd != null) {
                    lasd.refresh(allButN, br, step, bestStep, fwpb, kWithinPercentage);
                }
                br = br2;
            }
        }
        if (lasd != null) {
            lasd.refresh(allResults.getNBestResults(nSimultaneousTests), br, -1, bestStep, fwpb, kWithinPercentage);
        }
    } 

    public static void createNewMainDisplay() {
        final JFrame frame = new JFrame("lavcGenetic");
        final Map codecMap = new HashMap();
        lavcCodecData lcd;

        lcd = new lavcCodecData("lavc", "-lavcopts", buildLavcOpts(), "vpass", "vbitrate");
        codecMap.put("libavcodec", lcd);
        
        lcd = new lavcCodecData("xvid", "-xvidencopts", buildXvidencOpts(), "pass", "bitrate");
        codecMap.put("XviD", lcd);
        
        final JPanel mainDisplay = new lavcGeneticMainDisplay(frame, codecMap, "libavcodec");

        frame.getContentPane().setLayout(new BoxLayout(frame.getContentPane(), javax.swing.BoxLayout.Y_AXIS));
        frame.getContentPane().add(mainDisplay);
        frame.pack();
        frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
    }
}
