/* 
 * Copyright (C) 2004 Romain Dolbeau <romain@dolbeau.org>
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */

package lavcGenetic;


import java.util.*;
import java.io.*;

class lavcMencoderWrapperLocal implements lavcMencoderWrapper {
    private String mencoder = "p4xeon";
    private String mplayer = "mplayer";
    private int timeLimit;

    static private class keepAliveControl {
        Object o = new Object();
        boolean alive = true;

        void keepAlive() {
            synchronized (o) {
                alive = true;
            }
        }

        boolean isAlive() {
            boolean temp;
            synchronized (o) {
                temp = alive;
                alive = false;
            }
            return temp;
        }
    }

    lavcMencoderWrapperLocal(int t) {
        timeLimit = t;
    }

    lavcMencoderWrapperLocal(String me, String mp, int t) {
        mencoder = me;
        mplayer = mp;
        timeLimit = t;
    }

    static private double extractPSNRlavc(String line, String key) {
        // PSNR: Y:43.92, Cb:47.75, Cr:48.59, All:44.93
        double result = -1.;
        
        int start = line.indexOf(key);
        if (start != -1) {
            start += (key.length() + 1);
            int stop = line.indexOf(",", start);
            
            if (stop == -1)
                result = Double.parseDouble(line.substring(start));
            else
                result = Double.parseDouble(line.substring(start, stop));
        }
        return result;
    }

    static private double extractPSNRxvid(String line, String key) {
        //xvid: Average PSNR y : 47.47 dB, u : 49.73 dB, v : 50.70 dB, for 47 frames
        double result = -1.;
        int start = line.indexOf(key + " :");
        
        if (start != -1) {
            start += (key.length() + 3);
            int stop = line.indexOf("dB", start);
            result = Double.parseDouble(line.substring(start, stop -1));
        }
        return result;
        
    }

    static private double extractPosixTime(String line) {
        // user 0.00
        double result = -1.;

        int start = line.indexOf(" ");
        if (start != -1) {
            start++;
            result = Double.parseDouble(line.substring(start));
        }

        return result;
    }

    static private int extractFromPos(String line) {
        // Pos:  56.0s   2804f ( 0%) 762fps Trem:   7min 2526mb  A-V:0.053 [2617:448] A/Vms 24/0 D/B/S 0/0/1403
        int stop = line.indexOf("f");
        int start = stop - 1;
        while(line.charAt(start) != ' ')
            start--;
        start++;
        return (Integer.parseInt(line.substring(start,stop)));
    }
    
    public lavcResults dumpMoviePassX(lavcEncodeData led , int pass, int bitrate, String machine) {
        double Y = -1., Cb = -1., Cr = -1., All = -1.;
        double seconds = -1.;
        
        final String cmd =
            "/bin/nice -19 /usr/bin/time -p" +
            " " + mencoder +
            " " + led.source + 
            " -aid " + led.aid + 
            " " + led.videoFilter + 
            " -oac " + led.audioCodec + " " + led.audioOptions +
            " -ovc " + led.codecData.codec + " " + led.codecData.codecOpts + " " + led.getLacvOptsString() + ":" +
            led.codecData.bitrate + "=" + bitrate +
            (pass > 0 ? ":" + led.codecData.pass + "=" + pass : "") +
            " -o " + led.output;
        
        lavcGeneticMain.err_println("EXEC (" + machine + "): [pass " + pass + "] " + cmd);
        File workdir = null;
        try {
            workdir = new File(led.workdir);
            if (!workdir.exists()) {
                workdir.mkdirs();
            } else {
                if (pass < 2) {
                    File divx2passlog = new File(led.workdir + "/" + "divx2pass.log");
                    if (divx2passlog.exists())
                        divx2passlog.delete();
                    File xvidtwopassstats = new File(led.workdir + "/" + "xvid-twopass.stats");
                    if (xvidtwopassstats.exists())
                        xvidtwopassstats.delete();
                }
            }
        } catch (Exception e) { lavcGeneticMain.err_println("Exception " + e + " while creating File for " + led.workdir); }
        try {
            final Process theMenc = Runtime.getRuntime().exec(cmd, null, workdir);
            final keepAliveControl kac = new keepAliveControl();
            final Thread ourThread = Thread.currentThread();
            final InputStream theMencIn = theMenc.getInputStream();
            final InputStream theMencEr = theMenc.getErrorStream();
            final BufferedReader bufEr = new BufferedReader(new InputStreamReader(theMencEr));
            final BufferedReader bufIn = new BufferedReader(new InputStreamReader(theMencIn));
            final String machineCopy = machine;
            Timer safetyTimer = null;
            try {
                safetyTimer = new Timer(true);
                safetyTimer.schedule(new TimerTask() {
                        private int count = 0;
                        
                        public void run() {
                            boolean check = kac.isAlive();
                            if (!check) {
                                count ++;
                                lavcGeneticMain.err_println("WARNING (" + machineCopy + "): command " + cmd + " has timed out");
                                //ourThread.interrupt();
                                if (count >= 2) {
                                    lavcGeneticMain.err_println("->ERROR (" + machineCopy + "): command " + cmd + " has timed out *twice*");
                                    try {
                                        int i = theMenc.exitValue();
                                        lavcGeneticMain.err_println("--> ERROR (" + machineCopy + "): process has finished with status " + i);
                                    } catch (IllegalThreadStateException e) {
                                        lavcGeneticMain.err_println("--> ERROR (" + machineCopy + "): process has *not* finished, trying to destroy...");
                                        theMenc.destroy();
                                    } catch (Exception e) { lavcGeneticMain.err_println("---> ERROR (" + machineCopy + "): Exception " + e + " while trying to forcibly check/destroy the process..."); }
                                    try {
                                        lavcGeneticMain.err_println("--> ERROR (" + machineCopy + "): Trying to close all channels");
                                        bufIn.close();
                                        bufEr.close();
                                        theMencIn.close();
                                        theMencEr.close();
                                    } catch (Exception e) { lavcGeneticMain.err_println("---> ERROR (" + machineCopy + "): Exception " + e + " while trying to forcibly close the process channels..."); }
                                    lavcGeneticMain.err_println("WARNING (" + machineCopy + "): command " + cmd + " has timed out, I hope I have solved the problem..."); 
                                }
                            } else { count = 0; }
                        }
                    }, timeLimit * 1000 * 2, timeLimit * 1000);
            } catch (Exception e) { lavcGeneticMain.err_println("Exception " + e + " while creating Timer"); }
            
            
            String line = null;
            boolean done = false;
            kac.keepAlive();
            done = false;
            line = bufIn.readLine();
            while ((line != null) && (!done)) {
                kac.keepAlive();
                if (line.indexOf("PSNR:") > -1) { // libavcodec
                    Y = extractPSNRlavc(line, "Y");
                    Cb = extractPSNRlavc(line, "Cb");
                    Cr = extractPSNRlavc(line, "Cr");
                    All = extractPSNRlavc(line, "All");
                    done = true;
                }
                if ((line.indexOf("xvid:") > -1) && (line.indexOf("Average PSNR") > -1)) { // xvid
                    Y = extractPSNRxvid(line, "y");
                    Cb = extractPSNRxvid(line, "u");
                    Cr = extractPSNRxvid(line, "v");
                    All = (2 * Y + Cb + Cr) / 4.;
                    done = true;
                }
                if (line.indexOf("user") > -1) {
                    seconds = extractPosixTime(line);
                }
                /*
                  if (line.indexOf("Pos:") == 0) {
                  int lnum =  extractFromPos(line);
                  if ((lnum % 100) == 0)
                  lavcGeneticMain.err_println("Frame # " + lnum);
                  }
                */
//lavcGeneticMain.err_println("OUT: " + line);
                if (!done)
                    line = bufIn.readLine();
            }
            kac.keepAlive();
            done = false;
            line = bufEr.readLine();
            while ((line != null) && (!done)) {
                kac.keepAlive();
                if (line.indexOf("user") > -1) {
                    seconds = extractPosixTime(line);
                    done = true;
                }
                if (line.indexOf("Command terminated") >= 0) {
                    done = true;
                }
lavcGeneticMain.err_println("ERR: " + line);
                if (!done)
                    line = bufEr.readLine();
            }
            kac.keepAlive();
            int result = theMenc.waitFor();
            if (safetyTimer != null)
                safetyTimer.cancel();
            
            if (result != 0) {
                lavcGeneticMain.err_println("Command: \"" + cmd + "\" returned " + result);
            }
            
            bufIn.close();
            bufEr.close();
            theMencIn.close();
            theMencEr.close();
        } catch (Exception e) { lavcGeneticMain.err_println("Exception " + e + " while processing command \"" + cmd + "\""); }
        
        return new lavcResults(Y, Cb, Cr, All, seconds);
    }
}
