/* 
 * Copyright (C) 2004 Romain Dolbeau <romain@dolbeau.org>
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */

package lavcGenetic;

import java.util.*;

class lavcResultsPool {
    private Map allResults = null;
    private int invalidCount = 0;

    lavcResultsPool() {
        allResults = Collections.synchronizedMap(new HashMap());
    }

    lavcResultsPool(lavcResultsPool lrp) {
        synchronized (lrp.allResults) {
            allResults = Collections.synchronizedMap(new HashMap(lrp.allResults));
        }
    }

    boolean hasResult(lavcOptionList opt) {
        boolean result = false;
        synchronized (allResults) {
            if (allResults.containsKey(opt)) {
                result = true;
            }
        }
        return result;
    }

    lavcResults getResult(lavcOptionList opt) {
        lavcResults result = null;
        synchronized (allResults) {
            if (allResults.containsKey(opt)) {
                result = (lavcResults)allResults.get(opt);
            }
        }
        return result;
    }
    
    void addResult(lavcOptionList opt, lavcResults r) {
        if (hasResult(opt)) {
            lavcResults r2 = getResult(opt);
            lavcGeneticMain.err_println("ERROR !!!! *not* adding an already present result for " + opt + " new result: " + r + " old result: " + r2);
        } else {
            synchronized (allResults) {
                allResults.put(opt, r);
                if (!r.isValid())
                    invalidCount++;
            }
        }
    }

    private void addResultIgnoreDup(lavcOptionList opt, lavcResults r) {
        if (!hasResult(opt)) {
            synchronized (allResults) {
                allResults.put(opt, r);
                if (!r.isValid())
                    invalidCount++;
            }
        }
    }

    void mergeResults(lavcResultsPool lrp) {
        synchronized (lrp.allResults) {
            Iterator it = lrp.allResults.keySet().iterator();
            while (it.hasNext()) {
                lavcOptionList opt = (lavcOptionList)it.next();
                lavcResults result = (lavcResults)lrp.allResults.get(opt);
                addResultIgnoreDup(opt, result);
            }
        }
    }

    int getInvalidCount() {
        int temp = 0;
        synchronized (allResults) {
            temp = invalidCount;
        }
        return temp;
    }
    

    int numberOfResults() {
        int n = -1;

        synchronized (allResults) {
            n = allResults.size();
        }

        return n;
    }

    Map getAllResults() {
        Map copy;

        synchronized (allResults) {
            copy = new HashMap(allResults);
        }

        return copy;
    }

    lavcOptionList getBestResultKey() {
        lavcResults bestResult = null;
        lavcOptionList best = null;

        synchronized (allResults) {
            if (allResults.size() > 0) {
                Iterator it = allResults.keySet().iterator();
                best = (lavcOptionList)it.next();
                bestResult = (lavcResults)allResults.get(best);
                while (it.hasNext()) {
                    lavcOptionList temp = (lavcOptionList)it.next();
                    lavcResults tempResult = (lavcResults)allResults.get(temp);

                    if (tempResult.compareTo(bestResult) > 0) {
                        best = temp;
                        bestResult = tempResult;
                    }
                }
            }
        }
        return best;
    }

    lavcOptionList getWorstResultKey() {
        lavcResults worstResult = null;
        lavcOptionList worst = null;

        synchronized (allResults) {
            if (allResults.size() > 0) {
                Iterator it = allResults.keySet().iterator();
                do {
                    worst = (lavcOptionList)it.next();
                    worstResult = (lavcResults)allResults.get(worst);
                } while ((worst != null) && (it.hasNext()) && (!worstResult.isValid()));
                if (worst != null) {
                    while (it.hasNext()) {
                        lavcOptionList temp = (lavcOptionList)it.next();
                        lavcResults tempResult = (lavcResults)allResults.get(temp);
                        
                        if ((tempResult.compareTo(worstResult) < 0) &&
                            (tempResult.isValid())) {
                            worst = temp;
                            worstResult = tempResult;
                        }
                    }
                }
            }
        }
        return worst;
    }

    lavcResults getBestResult() {
        lavcOptionList best = getBestResultKey();
        lavcResults bestResult = null;
        synchronized (allResults) {
            bestResult = (lavcResults)allResults.get(best);
        }
        return bestResult;
    }

    private void removeResult(lavcOptionList l) {
        synchronized (allResults) {
            lavcResults result = (lavcResults)allResults.get(l);
            allResults.remove(l);
            if (!result.isValid())
                invalidCount--;
        }
    }

    private void removeAllInvalid() {
        synchronized (allResults) {
            Iterator it = allResults.keySet().iterator();
            while (it.hasNext()) {
                List l = (List)it.next();
                lavcResults r = (lavcResults)allResults.get(l);
                if (!r.isValid()) {
                    it.remove(); // don't use allResults.remove(l) to avoid destroying iterator
                }
            }
        }
    }

    lavcOptionList getFastestKey() {
        lavcResults bestResult = null;
        lavcOptionList best = null;

        synchronized (allResults) {
            if (allResults.size() > 0) {
                Iterator it = allResults.keySet().iterator();
                do {
                    best = (lavcOptionList)it.next();
                    bestResult = (lavcResults)allResults.get(best);
                } while ((best != null) && (it.hasNext()) && (!bestResult.isValid()));
                if (best != null) {
                    while (it.hasNext()) {
                        lavcOptionList temp = (lavcOptionList)it.next();
                        lavcResults tempResult = (lavcResults)allResults.get(temp);
                        
                        if ((tempResult.seconds < bestResult.seconds) &&
                            (tempResult.seconds > 0)) {
                            best = temp;
                            bestResult = tempResult;
                        }
                    }
                }
            }
        }
        return best;
    }

    List getNBestKeys(int n) {
        lavcResultsPool copy = new lavcResultsPool(this);
        List temp = new ArrayList();
        for (int i = 0 ; (i < n) && (copy.numberOfResults() > 0) ; i++) {
            lavcOptionList l = copy.getBestResultKey();
            copy.removeResult(l);
            temp.add(l);
        }
        return temp;
    } 

    lavcResultsPool getNBestResults(int n) {
        lavcResultsPool copy = new lavcResultsPool(this);
        lavcResultsPool temp = new lavcResultsPool();
        for (int i = 0 ; (i < n) && (copy.numberOfResults() > 0) ; i++) {
            lavcOptionList l = copy.getBestResultKey();
            lavcResults r = copy.getResult(l);
            copy.removeResult(l);
            temp.addResult(l, r);
        }
        return temp;
    }

    List getNWorstKeys(int n) {
        lavcResultsPool copy = new lavcResultsPool(this);
        List temp = new ArrayList();
        for (int i = 0 ; (i < n) && (copy.numberOfResults() > 0) ; i++) {
            lavcOptionList l = copy.getWorstResultKey();
            if (l != null) {
                copy.removeResult(l);
                temp.add(l);
            }
        }
        return temp;
    }

    lavcResultsPool getAllButHalfTheWorstN(int n, Random random) {
        lavcResultsPool copy = new lavcResultsPool(this);
        copy.removeAllInvalid();
        List worst = copy.getNWorstKeys(n);
        for (int i = 0; (i < n / 2) && (worst.size() > 0) ; i ++) {
            Object temp = worst.get(random.nextInt(worst.size()));
            worst.remove(temp);
        }
        Iterator it = worst.iterator();
        while (it.hasNext()) {
            copy.removeResult((lavcOptionList)it.next());
        }
        return copy;
    }

    lavcOptionList fasterWithinPercentBestKey(double p) {
        lavcResultsPool copy = new lavcResultsPool(this);
        lavcResultsPool best = new lavcResultsPool();
        lavcOptionList l = copy.getBestResultKey();
        lavcResults b = getResult(l);
        boolean done = false;

        while ((copy.numberOfResults() > 0) && !done) {
            lavcOptionList l2 = copy.getBestResultKey();
            lavcResults b2 = copy.getResult(l2);
            copy.removeResult(l2);
            if (b2.psnrAll >= (b.psnrAll * (1. - (p / 100.)))) {
                best.addResult(l2, b2);
            } else {
                done = true;
            }
        }

        return best.getFastestKey();
    }

    List getListOfKeys() {
        Set theSet;
        synchronized (allResults) {
            theSet = allResults.keySet();
        }
        return new ArrayList(theSet);
    }

    public String toString() {
        return "lavcResultsPool: " + allResults;
    }

    Map getStatisticsMap() {
        Map map = new HashMap();
        
        synchronized (allResults) {
            Iterator it = allResults.keySet().iterator();
            while (it.hasNext()) {
                lavcOptionList l = (lavcOptionList)it.next();
                Iterator it2 = l.iterator();
                while (it2.hasNext()) {
                    lavcOptionInstance loi = (lavcOptionInstance)it2.next();
                    if (map.containsKey(loi)) {
                        Integer old = (Integer)map.get(loi);
                        map.put(loi, new Integer(old.intValue() + 1));
                    } else {
                        map.put(loi, new Integer(1));
                    }
                }
            }
        }

        return map;
    }

    String optionStatistics(Set allOptions) {
        Map map = getStatisticsMap();
        double n = (double)numberOfResults();
        String temp = "";
        Iterator it = allOptions.iterator();
        while (it.hasNext()) {
            lavcOption opt = (lavcOption)it.next();
            temp = temp + "\t" + opt.getDescription() + ": ";
            Iterator it2 = opt.getAllVariants().iterator();
            while (it2.hasNext()) {
                lavcOptionInstance loi = (lavcOptionInstance)it2.next();
                Integer a;
                if (map.containsKey(loi))
                    a = (Integer)map.get(loi);
                else
                    a = new Integer(0);
                temp = temp + "\"" + loi.toString() + "\" = " + a + " (" + (a.doubleValue() * 100. / n) + " %) ";
            }
            temp = temp + "\n";
        }
        return temp;
    }
}
